pub type Reine = (i32, i32);

pub type BackTracker = Vec<Reine>;

use heuristics::{nb_blocked_lines, HeuristicsApproach};

pub fn can_be_put(r: Reine, reines: &[Reine]) -> bool {
    for o in reines {
        if r.0 == o.0 || r.1 == o.1 || (o.0 - r.0).abs() == (o.1 - r.1).abs() {
            return false;
        }
    }

    true
}

fn tracker_string(tracker: &[Reine]) -> String {
    let mut result = String::from("[");

    for (i, r) in tracker.iter().enumerate() {
        // close string at last element
        if i == tracker.len() - 1 {
            result.push_str(&format!("{:?}", r));
        } else {
            result.push_str(&format!("{:?}, ", r));
        }
    }

    result.push(']');
    result
}

pub fn nreines(n: usize, verbose: bool) -> usize {
    let mut tracker = BackTracker::new();

    fn f(n: usize, tracker: &mut BackTracker, verbose: bool) -> usize {
        let solution = if tracker.len() == n {
            if verbose {
                println!("Solution: {}\n", tracker_string(tracker));
            }
            1
        } else {
            let mut solution = 0;

            let len = tracker.len();

            if verbose && len > 0 {
                println!("{}{:?}", "  ".repeat(len - 1), tracker[len - 1]);
            }

            let row = len as i32;

            for i in 0..n {
                let reine: Reine = (row, i as i32);
                let ok = can_be_put(reine, tracker);

                if ok {
                    tracker.push(reine);
                    solution += f(n, tracker, verbose);
                }
            }

            solution
        };

        tracker.pop();

        solution
    }

    f(n, &mut tracker, verbose)
}

pub fn nreines_first_solution(n: usize, verbose: bool) -> bool {
    let mut tracker = BackTracker::new();

    fn f(n: usize, tracker: &mut BackTracker, verbose: bool) -> bool {
        if tracker.len() == n {
            if verbose {
                println!("Sol({:>2}): {}\n", n, tracker_string(tracker));
            }
            return true;
        } else {
            let row = tracker.len() as i32;

            for i in 0..n {
                let reine: Reine = (row, i as i32);
                let ok = can_be_put(reine, tracker);

                if ok {
                    tracker.push(reine);
                    if f(n, tracker, verbose) {
                        return true;
                    }
                }
            }
        };

        tracker.pop();

        false
    }

    f(n, &mut tracker, verbose)
}

pub fn nreines_first_solution_with_heuristics(
    n: usize,
    verbose: bool,
    approach: HeuristicsApproach,
) -> bool {
    let mut tracker = BackTracker::new();

    fn f(n: usize, tracker: &mut BackTracker, verbose: bool, approach: HeuristicsApproach) -> bool {
        //Terminal case: all queens have been put
        if tracker.len() == n {
            if verbose {
                println!("Sol({:>2}): {}\n", n, tracker_string(tracker));
            }
            return true;
        }
        //Recursive case, when there's still queens to put
        else {
            let row = tracker.len() as i32;

            //Maps each potential queen to the number of blocks
            //that would be occupied if she was put
            let mut columns_heuristics: Vec<_> = (0..n)
                .into_iter()
                .map(|i| (row, i as i32))
                .filter(|r| can_be_put(*r, tracker))
                .map(|r| (r, nb_blocked_lines(n, r, tracker)))
                .collect();

            //Sorts by number of blocked lines
            //based on given approach
            columns_heuristics.sort_by_key(|h| match approach {
                HeuristicsApproach::Opportunist => (h.1 as i32),
                HeuristicsApproach::Failing => -(h.1 as i32),
            });

            for (reine, _) in columns_heuristics {
                tracker.push(reine);

                if f(n, tracker, verbose, approach) {
                    return true;
                }
            }
        };

        tracker.pop();

        false
    }

    f(n, &mut tracker, verbose, approach)
}
